"""
This testcase is used to test the REST API  in api/machines.py
to fetch machine information and to attach token to machines
"""
from .base import MyApiTestCase
from privacyidea.lib.token import init_token

HOSTSFILE = "tests/testdata/hosts"

SSHKEY = "ssh-rsa " \
         "AAAAB3NzaC1yc2EAAAADAQABAAACAQDJy0rLoxqc8SsY8DVAFijMsQyCv" \
         "hBu4K40hdZOacXK4O6OgnacnSKN56MP6pzz2+4svzvDzwvkFsvf34pbsgD" \
         "F67PPSCsimmjEQjf0UfamBKh0cl181CbPYsph3UTBOCgHh3FFDXBduPK4DQz" \
         "EVQpmqe80h+lsvQ81qPYagbRW6fpd0uWn9H7a/qiLQZsiKLL07HGB+NwWue4os" \
         "0r9s4qxeG76K6QM7nZKyC0KRAz7CjAf+0X7YzCOu2pzyxVdj/T+KArFcMmq8V" \
         "dz24mhcFFXTzU3wveas1A9rwamYWB+Spuohh/OrK3wDsrryStKQv7yofgnPMs" \
         "TdaL7XxyQVPCmh2jVl5ro9BPIjTXsre9EUxZYFVr3EIECRDNWy3xEnUHk7Rzs" \
         "734Rp6XxGSzcSLSju8/MBzUVe35iXfXDRcqTcoA0700pIb1ANYrPUO8Up05v4" \
         "EjIyBeU61b4ilJ3PNcEVld6FHwP3Z7F068ef4DXEC/d7pibrp4Up61WYQIXV/" \
         "utDt3NDg/Zf3iqoYcJNM/zIZx2j1kQQwqtnbGqxJMrL6LtClmeWteR4420uZx" \
         "afLE9AtAL4nnMPuubC87L0wJ88un9teza/N02KJMHy01Yz3iJKt3Ou9eV6kqO" \
         "ei3kvLs5dXmriTHp6g9whtnN6/Liv9SzZPJTs8YfThi34Wccrw== " \
         "NetKnights GmbH"
SSHKEY_ecdsa = "ecdsa-sha2-nistp256 AAAAE2VjZHNhLXNoYTItbmlzdHAyNTYAAAAIbmlzd" \
               "HAyNTYAAABBBHGCdIk0pO1HFr/mF4oLb43ZRyQJ4K7ICLrAhAiQERVa0tUvyY5TE" \
               "zurWTqxSMx203rY77t6xnHLZBMPPpv8rk0= cornelius@puck"
OTPKEY = "3132333435363738393031323334353637383930"


class APIMachinesServiceIDTestCase(MyApiTestCase):

    serial1 = "SSHKEY1"
    serial2 = "SSHKEY2"
    serviceID1 = "webserver"
    serviceID2 = "mailserver"

    def test_01_create_sshkeys(self):
        # create two tokens
        token_obj = init_token({"serial": self.serial1, "type": "sshkey",
                                "sshkey": SSHKEY})
        self.assertEqual(token_obj.type, "sshkey")
        token_obj = init_token({"serial": self.serial2, "type": "sshkey",
                                "sshkey": SSHKEY_ecdsa})
        self.assertEqual(token_obj.type, "sshkey")

    def test_02_attach_tokens(self):
        # Do the following token attachemends:
        # * S1: webserver and mailserver
        # * S2: only mailserver
        # Attach token S1 to webserver
        with self.app.test_request_context('/machine/token',
                                           method='POST',
                                           data={"serial": self.serial1,
                                                 "application": "ssh",
                                                 "user": "root",
                                                 "service_id": self.serviceID1},
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            result = res.json.get("result")
            self.assertEqual(result["status"], True)
            self.assertTrue(result["value"] >= 1)

        # Attach S1 to mailserver
        with self.app.test_request_context('/machine/token',
                                           method='POST',
                                           data={"serial": self.serial1,
                                                 "application": "ssh",
                                                 "user": "root",
                                                 "service_id": self.serviceID2},
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            result = res.json.get("result")
            self.assertEqual(result["status"], True)
            self.assertTrue(result["value"] >= 1)

        # Attach S2 to mailserver
        with self.app.test_request_context('/machine/token',
                                           method='POST',
                                           data={"serial": self.serial2,
                                                 "application": "ssh",
                                                 "user": "root",
                                                 "service_id": self.serviceID2},
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            result = res.json.get("result")
            self.assertEqual(result["status"], True)
            self.assertTrue(result["value"] >= 1)

    def test_03_get_service_ids(self):
        # Get all machinetokens
        with self.app.test_request_context('/machine/token',
                                           method='GET',
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            result = res.json.get("result")
            self.assertEqual(result["status"], True)
            value = result.get("value")
            self.assertEqual(len(value), 3)
            self.assertEqual(value[0]["application"], "ssh")
            self.assertEqual(value[1]["application"], "ssh")
            self.assertEqual(value[2]["application"], "ssh")

        # Get tokens for service_id self.serviceID1
        with self.app.test_request_context('/machine/token?service_id={0!s}'.format(self.serviceID1),
                                           method='GET',
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            result = res.json.get("result")
            self.assertEqual(result["status"], True)
            value = result.get("value")
            self.assertEqual(len(value), 1)
            self.assertEqual(value[0]["application"], "ssh")
            self.assertEqual(value[0].get("options").get("service_id"), self.serviceID1)
            self.assertEqual(value[0].get("serial"), self.serial1)

        # Get token for service_id self.serviceID2
        with self.app.test_request_context('/machine/token?service_id={0!s}'.format(self.serviceID2),
                                           method='GET',
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            result = res.json.get("result")
            self.assertEqual(result["status"], True)
            value = result.get("value")
            self.assertEqual(len(value), 2)
            self.assertEqual(value[0]["application"], "ssh")
            self.assertEqual(value[0].get("options").get("service_id"), self.serviceID2)
            self.assertEqual(value[0].get("serial"), self.serial1)
            self.assertEqual(value[1]["application"], "ssh")
            self.assertEqual(value[1].get("options").get("service_id"), self.serviceID2)
            self.assertEqual(value[1].get("serial"), self.serial2)

        # combine filter and get service_id self.serviceID2 for serial1
        with self.app.test_request_context('/machine/token?service_id={0!s}&serial={1!s}'.format(
                self.serviceID2, self.serial1),
                                           method='GET',
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            result = res.json.get("result")
            self.assertEqual(result["status"], True)
            value = result.get("value")
            self.assertEqual(len(value), 1)
            self.assertEqual(value[0]["application"], "ssh")
            self.assertEqual(value[0].get("options").get("service_id"), self.serviceID2)
            self.assertEqual(value[0].get("serial"), self.serial1)

        # Get token for service_id self.serviceID2 and the correct application
        with self.app.test_request_context('/machine/token?service_id={0!s}&application=ssh'.format(self.serviceID2),
                                           method='GET',
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            result = res.json.get("result")
            self.assertEqual(result["status"], True)
            value = result.get("value")
            self.assertEqual(len(value), 2)
            self.assertEqual(value[0]["application"], "ssh")
            self.assertEqual(value[0].get("options").get("service_id"), self.serviceID2)
            self.assertEqual(value[0].get("serial"), self.serial1)
            self.assertEqual(value[1]["application"], "ssh")
            self.assertEqual(value[1].get("options").get("service_id"), self.serviceID2)
            self.assertEqual(value[1].get("serial"), self.serial2)

        # Get token for service_id self.serviceID2 and the wrong application
        with self.app.test_request_context(
                '/machine/token?service_id={0!s}&application=openssh'.format(self.serviceID2),
                method='GET',
                headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            result = res.json.get("result")
            self.assertEqual(result["status"], True)
            value = result.get("value")
            self.assertEqual(len(value), 0)

    def test_04_get_service_id_for_different_users(self):
        # Add another SSH key for user admin to self.serviceID2
        with self.app.test_request_context('/machine/token',
                                           method='POST',
                                           data={"serial": self.serial1,
                                                 "application": "ssh",
                                                 "user": "admin",
                                                 "service_id": self.serviceID2},
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            result = res.json.get("result")
            self.assertEqual(result["status"], True)
            self.assertTrue(result["value"] >= 1)

        # Get token for service_id self.serviceID2 and the application=ssh and the user=root
        with self.app.test_request_context(
                '/machine/token?service_id={0!s}&application=ssh&user=root'.format(self.serviceID2),
                method='GET',
                headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            result = res.json.get("result")
            self.assertEqual(result["status"], True)
            value = result.get("value")
            self.assertEqual(len(value), 2)
            self.assertEqual(value[0]["application"], "ssh")
            self.assertEqual(value[0].get("options").get("service_id"), self.serviceID2)
            self.assertEqual(value[0].get("serial"), self.serial1)
            self.assertEqual(value[1]["application"], "ssh")
            self.assertEqual(value[1].get("options").get("service_id"), self.serviceID2)
            self.assertEqual(value[1].get("serial"), self.serial2)

        # Get token for service_id self.serviceID2 and the application=ssh and the user=admin
        with self.app.test_request_context(
                '/machine/token?service_id={0!s}&application=ssh&user=admin'.format(self.serviceID2),
                method='GET',
                headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            result = res.json.get("result")
            self.assertEqual(result["status"], True)
            value = result.get("value")
            self.assertEqual(len(value), 1)
            self.assertEqual(value[0]["application"], "ssh")
            self.assertEqual(value[0].get("options").get("service_id"), self.serviceID2)
            self.assertEqual(value[0].get("serial"), self.serial1)

        # test filter and wildcards

        # Find all tokens attached to a *server
        with self.app.test_request_context(
                '/machine/token?service_id=*server*&application=ssh',
                method='GET',
                headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            result = res.json.get("result")
            self.assertEqual(result["status"], True)
            value = result.get("value")
            self.assertEqual(len(value), 4)
            self.assertEqual(value[0]["application"], "ssh")

        # Find all tokens attached to the user *root*
        with self.app.test_request_context(
                '/machine/token?application=ssh&user=*oot',
                method='GET',
                headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            result = res.json.get("result")
            self.assertEqual(result["status"], True)
            value = result.get("value")
            # 3 tokens are attached via user "root"
            self.assertEqual(len(value), 3)
            self.assertEqual(value[0]["application"], "ssh")

        # Find all tokens attached to the user *adm*
        with self.app.test_request_context(
                '/machine/token?application=ssh&user=*adm*',
                method='GET',
                headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            result = res.json.get("result")
            self.assertEqual(result["status"], True)
            value = result.get("value")
            # One token is attached via user "admin"
            self.assertEqual(len(value), 1)
            self.assertEqual(value[0]["application"], "ssh")

        # Filter vor *KEY1
        with self.app.test_request_context(
                '/machine/token?application=ssh&serial=*KEY1',
                method='GET',
                headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            result = res.json.get("result")
            self.assertEqual(result["status"], True)
            value = result.get("value")
            # One token is attached via user "admin"
            self.assertEqual(len(value), 3)
            for v in value:
                # We only get SSHKEY1
                self.assertEqual(v.get("serial"), self.serial1)

        # sort by service_id
        with self.app.test_request_context(
                '/machine/token?application=ssh&serial=*KEY1&sortby=service_id',
                method='GET',
                headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            result = res.json.get("result")
            self.assertEqual(result["status"], True)
            value = result.get("value")
            # One token is attached via user "admin"
            self.assertEqual(len(value), 3)
            self.assertEqual(value[0].get("options").get("user"), "root")
            self.assertEqual(value[1].get("options").get("user"), "admin")
            self.assertEqual(value[2].get("options").get("user"), "root")



#  2015-04-08 Cornelius Kölbel <cornelius.koelbel@netknights.it>
#             Add options ROUNDS to avoid timeouts during OTP hash calculation
#  2015-04-03 Cornelius Kölbel <cornelius.koelbel@netknights.it>
#             Use pbkdf2 for OTP hashing
#  2015-03-13 Cornelius Kölbel, <cornelius@privacyidea.org>
#             initial writeup
#
# License:  AGPLv3
#  contact:  http://www.privacyidea.org
#
# This code is free software; you can redistribute it and/or
# modify it under the terms of the GNU AFFERO GENERAL PUBLIC LICENSE
# License as published by the Free Software Foundation; either
# version 3 of the License, or any later version.
#
# This code is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU AFFERO GENERAL PUBLIC LICENSE for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import logging

from passlib.hash import pbkdf2_sha512

from privacyidea.lib.applications import MachineApplicationBase
from privacyidea.lib.config import get_prepend_pin
from privacyidea.lib.crypto import geturandom
from privacyidea.lib.error import ValidateError, ParameterError
from privacyidea.lib.policy import TYPE
from privacyidea.lib.token import get_one_token
from privacyidea.lib.utils import get_computer_name_from_user_agent

log = logging.getLogger(__name__)
ROUNDS = 6549
REFILLTOKEN_LENGTH = 40


class MachineApplication(MachineApplicationBase):
    """
    This is the application for Offline authentication with PAM or
    the privacyIDEA credential provider.

    The machine application returns a list of salted OTP hashes to be used with
    offline authentication. The token then is disabled, so that it can not
    be used for online authentication anymore, to avoid reusing a fished OTP
    value.

    The server stores the information, which OTP values were issued.

    options:
      * user: a username.
      * count: is the number of OTP values returned

    """
    application_name = "offline"

    @staticmethod
    def generate_new_refilltoken(token, user_agent=None):
        """
        Generate new refill token and store it in the tokeninfo of the token.
        :param token: token in question
        :param user_agent: name of the machine, taken from the user-agent header
        :return: a string
        """
        log.debug(f"Generating refilltoken for {token.get_serial()} with user agent {user_agent}")

        # If the token is a FIDO2 token, we need to store the machine name with the refill token, because
        # the token can be on multiple machines, which need to be managed separately
        # TODO improve where the information about offline capabilities is stored
        if token.type.lower() in ["webauthn", "passkey"]:
            computer_name = get_computer_name_from_user_agent(user_agent)
            if not computer_name:
                log.warning(f"No computer name found in user agent {user_agent}. Unable to generate refill token.")
                raise ParameterError("Unable to generate refill token without computer name.")
            key = "refilltoken_" + computer_name
        else:
            key = "refilltoken"

        new_refilltoken = geturandom(REFILLTOKEN_LENGTH, hex=True)
        token.add_tokeninfo(key, new_refilltoken)
        return new_refilltoken

    @staticmethod
    def get_offline_otps(token, otppin, amount, rounds=ROUNDS):
        """
        Retrieve the desired number of passwords (= PIN + OTP), hash them
        and return them in a dictionary. Increase the token counter.

        :param token: token in question
        :param otppin: The OTP PIN to prepend in the passwords. The PIN is not validated!
        :param amount: Number of OTP values (non-negative!)
        :param rounds: Number of PBKDF2 rounds
        :return: dictionary
        """
        if amount < 0:
            raise ParameterError("Invalid refill amount: {!r}".format(amount))
        (res, err, otp_dict) = token.get_multi_otp(count=amount, counter_index=True)
        otps = otp_dict.get("otp")
        prepend_pin = get_prepend_pin()
        for key, otp in otps.items():
            # Return the hash of OTP PIN and OTP values
            otp_with_pin = otppin + otp if prepend_pin else otp + otppin
            otps[key] = pbkdf2_sha512.using(rounds=rounds, salt_size=10).hash(otp_with_pin)
        # We do not disable the token, so if all offline OTP values
        # are used, the token can be used to authenticate online again.
        # token_obj.enable(False)
        # increase the counter by the consumed values and
        # also store it in tokeninfo.
        token.inc_otp_counter(increment=amount)

        return otps

    @staticmethod
    def get_refill(token, password, options=None):
        """
        Returns new authentication OTPs to refill the client

        To do so we also verify the password, which may consist of PIN + OTP.

        :param token: Token object
        :param password: PIN + OTP
        :param options: dict that might contain "count" and "rounds"
        :return: a dictionary of auth items
        """
        options = options or {}
        otps = {}
        if token.type.lower() == "hotp":
            count = int(options.get("count", 100))
            rounds = int(options.get("rounds", ROUNDS))
            _r, otppin, otpval = token.split_pin_pass(password)
            if not _r:
                raise ParameterError("Could not split password")
            current_token_counter = token.token.count
            first_offline_counter = current_token_counter - count
            if first_offline_counter < 0:
                first_offline_counter = 0
            # find the value in the offline OTP values! This resets the token.count!
            matching_count = token.check_otp(otpval, first_offline_counter, count)
            token.set_otp_count(current_token_counter)
            # Raise an exception *after* we reset the token counter
            if matching_count < 0:
                raise ValidateError("You provided a wrong OTP value.")
            # We have to add 1 here: Assume *first_offline_counter* is the counter value of the first offline OTP
            # we sent to the client. Assume the client then requests a refill with that exact OTP value.
            # Then, we need to respond with a refill of one OTP value, as the client has consumed one OTP value.
            counter_diff = matching_count - first_offline_counter + 1
            otps = MachineApplication.get_offline_otps(token, otppin, counter_diff, rounds)
            token.add_tokeninfo(key="offline_counter",
                                value=count)
        elif token.type.lower() in ["webauthn", "passkey"]:
            pass
        return otps

    @staticmethod
    def get_authentication_item(token_type, serial, challenge=None, options=None, filter_param=None, user_agent=None):
        """
        :param token_type: the type of the token. At the moment
                           we support HOTP and WebAuthn/Passkey tokens.
        :param serial:     the serial number of the token.
        :param challenge:  this can contain the password (otp pin + otp value)
                           so that we can put the OTP PIN into the hashed response.
        :type challenge: basestring
        :param options: options
        :param filter_param: parameters
        :param user_agent: The user agent of the request
        :return auth_item: A list of hashed OTP values or pubKey, rpId and credentialId for WebAuthn token
        """
        ret = {}
        options = options or {}
        password = challenge
        token_type = token_type.lower()
        # TODO improve where the information about offline capabilities is stored
        if token_type in ["hotp", "webauthn", "passkey"]:
            token = get_one_token(serial=serial)
            user = token.user
            if user:
                user_info = user.info
                if "username" in user_info:
                    ret["user"] = ret["username"] = user_info.get("username")

            ret["refilltoken"] = MachineApplication.generate_new_refilltoken(token, user_agent)

            # Gather offline data depending on the token type
            if token_type == "webauthn":
                # return the pubKey, rpId and the credentialId (contained in the otpkey) to allow the machine to
                # verify the WebAuthn assertions signed with the token
                ret["response"] = {
                    "pubKey": token.get_tokeninfo("pubKey"),
                    "credentialId": token.decrypt_otpkey(),
                    "rpId": token.get_tokeninfo("relying_party_id")
                }
            elif token_type == "hotp":
                if password:
                    success, otppin, _ = token.split_pin_pass(password)
                    if not success:
                        raise ParameterError("Could not split password")
                else:
                    otppin = ""

                ret["response"] = MachineApplication.get_offline_otps(token,
                                                                      otppin,
                                                                      int(options.get("count", 100)),
                                                                      int(options.get("rounds", ROUNDS)))
            elif token_type == "passkey":
                ret["response"] = {
                    "pubKey": token.get_tokeninfo("public_key"),
                    "rpId": token.get_tokeninfo("relying_party_id"),
                    "credentialId": token.token.get_otpkey().getKey().decode("utf-8")
                }
        else:
            log.info(f"Token {serial} of type {token_type} is not supported by OFFLINE application module")

        return ret

    @staticmethod
    def get_options():
        """
        Returns a dictionary with a list of required and optional options
        """
        options = {
            "hotp": {
                "count": {"type": TYPE.STRING},
                "rounds": {"type": TYPE.STRING}
            },
            "webauthn": {},
            "passkey": {}
        }
        return options



#  privacyIDEA is a fork of LinOTP
#  May 08, 2014 Cornelius Kölbel
#  License:  AGPLv3
#  contact:  http://www.privacyidea.org
#
#  2018-01-21 Cornelius Kölbel <cornelius.koelbel@netknights.it>
#             Add tokenkind
#  2015-01-28 Rewrite for migration to flask
#             Cornelius Kölbel <cornelius@privacyidea.org>
#
#
#  Copyright (C) 2010 - 2014 LSE Leading Security Experts GmbH
#  License:  LSE
#  contact:  http://www.linotp.org
#            http://www.lsexperts.de
#            linotp@lsexperts.de
#
# This code is free software; you can redistribute it and/or
# modify it under the terms of the GNU AFFERO GENERAL PUBLIC LICENSE
# License as published by the Free Software Foundation; either
# version 3 of the License, or any later version.
#
# This code is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU AFFERO GENERAL PUBLIC LICENSE for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
__doc__ = """This is the implementation of the remote token. The remote token
forwards an authentication request to another privacyidea server.

To do this it uses the parameters remote.server_id, remote.realm,
remote.resolver, remote.user or remote.serial.
The parameter remote.local_checkpin determines, whether the PIN should be
checked locally or remotely.

The code is tested in tests/test_lib_tokens_remote
"""

import logging
import traceback
import requests
from privacyidea.lib.utils import is_true
from privacyidea.lib.decorators import check_token_locked
from privacyidea.lib.config import get_from_config
from privacyidea.api.lib.utils import getParam
from privacyidea.lib.log import log_with
from privacyidea.lib.policydecorators import challenge_response_allowed
from privacyidea.lib.tokenclass import TokenClass, TOKENKIND, AUTHENTICATIONMODE
from privacyidea.lib.privacyideaserver import get_privacyideaserver
from privacyidea.lib import _
from privacyidea.lib.policy import SCOPE, ACTION, GROUP

optional = True
required = False

log = logging.getLogger(__name__)

###############################################


class RemoteTokenClass(TokenClass):
    """
    The Remote token forwards an authentication request to another privacyIDEA
    server. The request can be forwarded to a user on the other server or to
    a serial number on the other server. The PIN can be checked on the local
    privacyIDEA server or on the remote server.

    Using the Remote token you can assign one physical token to many
    different users.
    """
    mode = [AUTHENTICATIONMODE.AUTHENTICATE, AUTHENTICATIONMODE.CHALLENGE]

    def __init__(self, db_token):
        """
        constructor - create a token class object with it's db token binding

        :param aToken: the db bound token
        """
        TokenClass.__init__(self, db_token)
        self.set_type("remote")

    @staticmethod
    def get_class_type():
        """
        return the class type identifier
        """
        return "remote"

    @staticmethod
    def get_class_prefix():
        """
        return the token type prefix
        """
        return "PIRE"

    @staticmethod
    @log_with(log)
    def get_class_info(key=None, ret='all'):
        """
        :param key: subsection identifier
        :type key: string
        :param ret: default return value, if nothing is found
        :type ret: user defined
        :return: subsection if key exists or user defined
        :rtype: dict or string
        """
        res = {'type': 'remote',
               'title': 'Remote Token',
               'description': _('Remote Token: Forward authentication request '
                                'to another server.'),
               'user': [],
               # This tokentype is enrollable in the UI for...
               'ui_enroll': ["admin"],
               'policy': {
                   SCOPE.ENROLL: {
                       ACTION.MAXTOKENUSER: {
                           'type': 'int',
                           'desc': _("The user may only have this maximum number of remote tokens assigned."),
                           'group': GROUP.TOKEN
                       },
                       ACTION.MAXACTIVETOKENUSER: {
                           'type': 'int',
                           'desc': _(
                               "The user may only have this maximum number of active remote tokens assigned."),
                           'group': GROUP.TOKEN
                       }
                   }
               },
               }

        if key:
            ret = res.get(key, {})
        else:
            if ret == 'all':
                ret = res
        return ret

    def update(self, param):
        """
        second phase of the init process - updates parameters

        :param param: the request parameters
        :return: - nothing -
        """
        # if another OTP length would be specified in /admin/init this would
        # be overwritten by the parent class, which is ok.
        self.set_otplen(6)
        TokenClass.update(self, param)

        remote_server_id = getParam(param, "remote.server_id", optional)

        if remote_server_id is not None:
            if get_privacyideaserver(id=remote_server_id):
                self.add_tokeninfo("remote.server_id", remote_server_id)
        else:
            remoteServer = getParam(param, "remote.server", required)
            self.add_tokeninfo("remote.server", remoteServer)
            log.warning("Using remote.server for remote tokens is deprecated. Use remote.server_id!")

        val = getParam(param, "remote.local_checkpin", optional) or 0
        self.add_tokeninfo("remote.local_checkpin", val)

        for key in ["remote.serial", "remote.user", "remote.path",
                    "remote.realm", "remote.resolver"]:
            val = getParam(param, key, optional)
            if val is not None:
                self.add_tokeninfo(key, val)

        self.add_tokeninfo("tokenkind", TOKENKIND.VIRTUAL)

    @property
    def check_pin_local(self):
        """
        lookup if pin should be checked locally or on remote host

        :return: bool
        """
        local_check = is_true(self.get_tokeninfo("remote.local_checkpin"))
        log.debug(" local checking pin? {0!r}".format(local_check))

        return local_check

    @log_with(log)
    @check_token_locked
    def authenticate(self, passw, user=None, options=None):
        """
        do the authentication on base of password / otp and user and
        options, the request parameters.

        Here we contact the other privacyIDEA server to validate the OtpVal.

        :param passw: the password / otp
        :param user: the requesting user
        :param options: the additional request parameters

        :return: tuple of (success, otp_count - 0 or -1, reply)

        """
        res = False
        otp_counter = -1
        reply = None
        otpval = passw

        # should we check the pin localy?
        if self.check_pin_local:
            (_res, pin, otpval) = self.split_pin_pass(passw, user,
                                                      options=options)

            if not TokenClass.check_pin(self, pin, user=user, options=options):
                return False, otp_counter, {'message': "Wrong PIN"}

        otp_count = self.check_otp(otpval, options=options)
        if otp_count >= 0:
            res = True
            reply = {'message': 'matching 1 tokens',
                     'serial': self.get_serial(),
                     'type': self.get_tokentype()}
        else:
            reply = {'message': 'remote side denied access'}

        return res, otp_count, reply

    @check_token_locked
    def check_otp(self, otpval, counter=None, window=None, options=None):
        """
        run the http request against the remote host

        :param otpval: the OTP value
        :param counter: The counter for counter based otp values
        :type counter: int
        :param window: a counter window
        :type counter: int
        :param options: additional token specific options
        :type options: dict
        :return: counter of the matching OTP value.
        :rtype: int
        """
        otp_count = -1
        pi_server_obj = None
        remoteServer = None

        remote_server_id = self.get_tokeninfo("remote.server_id")
        if remote_server_id:
            pi_server_obj = get_privacyideaserver(id=int(remote_server_id))
        else:
            # Deprecated
            remoteServer = self.get_tokeninfo("remote.server") or ""
            log.warning("Using remote.server for remote tokens is deprecated. Use remote.server_id!")
            ssl_verify = get_from_config("remote.verify_ssl_certificate",
                                         False, return_bool=True) or False
            # in preparation of the ability to relocate privacyidea urls,
            # we introduce the remote url path
            remotePath = self.get_tokeninfo("remote.path") or ""
            remotePath = remotePath.strip()
            remotePath = remotePath or "/validate/check"

        remoteSerial = self.get_tokeninfo("remote.serial") or ""
        remoteUser = self.get_tokeninfo("remote.user") or ""
        remoteRealm = self.get_tokeninfo("remote.realm") or ""
        remoteResolver = self.get_tokeninfo("remote.resolver") or ""

        # here we also need to check for remote.user and so on....
        log.debug("checking OTP len:%r remotely on server: %r,"
                  " serial: %r, user: %r" %
                  (len(otpval), remoteServer, remoteSerial, remoteUser))
        params = {}

        if remoteSerial:
            params['serial'] = remoteSerial
        elif remoteUser:
            params['user'] = remoteUser
            params['realm'] = remoteRealm
            params['resolver'] = remoteResolver
        else:
            log.warning("The remote token does neither contain a "
                        "remote.serial nor a remote.user.")
            return otp_count

        try:
            if remoteServer:
                # Deprecated
                params['pass'] = otpval
                request_url = "{0!s}{1!s}".format(remoteServer, remotePath)
                r = requests.post(request_url, data=params, verify=ssl_verify, timeout=60)
            elif pi_server_obj:
                r = pi_server_obj.validate_check(remoteUser, otpval,
                                                 serial=remoteSerial,
                                                 realm=remoteRealm,
                                                 resolver=remoteResolver)

            if r.status_code == requests.codes.ok:
                response = r.json()
                result = response.get("result")
                if result.get("value"):
                    otp_count = 1
                    # Add the serial of the used remote token in the tokeninfo parameters
                    self.add_tokeninfo("last_matching_remote_serial", response.get("detail", {}).get("serial"))

        except Exception as exx:  # pragma: no cover
            log.error("Error getting response from "
                      "remote Server (%r): %r" % (request_url, exx))
            log.debug("{0!s}".format(traceback.format_exc()))

        return otp_count

    @log_with(log)
    @challenge_response_allowed
    def is_challenge_request(self, passw, user=None, options=None):
        """
        This method checks, if this is a request, that triggers a challenge.
        It depends on the way, the pin is checked - either locally or remote

        :param passw: password, which might be pin or pin+otp
        :type passw: string
        :param user: The user from the authentication request
        :type user: User object
        :param options: dictionary of additional request parameters
        :type options: dict

        :return: true or false
        """

        request_is_valid = False

        if self.check_pin_local:
            pin_match = self.check_pin(passw, user=user, options=options)
            if pin_match is True:
                request_is_valid = True

        return request_is_valid



"""v3.10: Add a table for policy descriptions

Revision ID: 5741e5dac477
Revises: e3a64b4ca634
Create Date: 2024-02-23 11:06:41.729152

"""
from alembic import op, context
import sqlalchemy as sa
from sqlalchemy.exc import OperationalError, ProgrammingError
from sqlalchemy.schema import Sequence, CreateSequence, DropSequence

# revision identifiers, used by Alembic.
revision = '5741e5dac477'
down_revision = 'db6b2ef8100f'


def upgrade():
    try:
        seq = Sequence('description_seq')
        try:
            if context.get_context().dialect.supports_sequences:
                op.execute(CreateSequence(seq))
        except (OperationalError, ProgrammingError) as exx:
            if "already exists" in str(exx.orig).lower():
                print(f"Ok, sequence '{seq}' already exists.")
            else:
                raise
        except Exception as _exx:
            print(f"Could not create sequence '{seq}'!")
            raise

        op.create_table('description',
                        sa.Column('id', sa.Integer(), seq, nullable=False),
                        sa.Column('object_type', sa.Unicode(length=64), nullable=False),
                        sa.Column('last_update', sa.DateTime),
                        sa.Column('description', sa.UnicodeText()),
                        sa.Column('object_id', sa.Integer(), nullable=False),
                        sa.ForeignKeyConstraint(('object_id',), ['policy.id'], ),
                        sa.PrimaryKeyConstraint('id'),
                        mysql_row_format='DYNAMIC'
                        )
    except (OperationalError, ProgrammingError) as exx:
        if "already exists" in str(exx.orig).lower():
            print("Ok, table 'description' already exists.")
        else:
            raise
    except Exception as _exx:
        print("Could not add table 'description'!")
        raise


def downgrade():
    op.drop_table('description')
    seq = Sequence('description_seq')
    if context.get_context().dialect.supports_sequences:
        op.execute(DropSequence(seq))



#  2018-11-22 Initial create
#             Cornelius Kölbel <cornelius.koelbel@netknights.it>
#
#  License:  AGPLv3
#  contact:  http://www.privacyidea.org
#
# This code is free software; you can redistribute it and/or
# modify it under the terms of the GNU AFFERO GENERAL PUBLIC LICENSE
# License as published by the Free Software Foundation; either
# version 3 of the License, or any later version.
#
# This code is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU AFFERO GENERAL PUBLIC LICENSE for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
__doc__ = """This module writes statistics data to the SQL database table "monitoringstats".
"""
import logging
from privacyidea.lib.monitoringmodules.base import Monitoring as MonitoringBase
from privacyidea.lib.pooling import get_engine
from privacyidea.lib.utils import censor_connect_string, convert_timestamp_to_utc
from privacyidea.lib.lifecycle import register_finalizer
from sqlalchemy import MetaData
from sqlalchemy import and_
from privacyidea.models import MonitoringStats
from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker, scoped_session
import traceback
from dateutil.tz import tzutc

log = logging.getLogger(__name__)

metadata = MetaData()


class Monitoring(MonitoringBase):

    def __init__(self, config=None):
        self.name = "sqlstats"
        self.config = config or {}
        self.engine = get_engine(self.name, self._create_engine)
        # create a configured "Session" class. ``scoped_session`` is not
        # necessary because we do not share session objects among threads.
        # We use it anyway as a safety measure.
        Session = scoped_session(sessionmaker(bind=self.engine))
        self.session = Session()
        # Ensure that the connection gets returned to the pool when the request has
        # been handled. This may close an already-closed session, but this is not a problem.
        register_finalizer(self.session.close)
        self.session._model_changes = {}

    def _create_engine(self):
        """
        :return: a new SQLAlchemy engine connecting to the database specified in PI_MONITORING_SQL_URI.
        """
        # an Engine, which the Session will use for connection
        # resources
        connect_string = self.config.get("PI_MONITORING_SQL_URI", self.config.get(
            "SQLALCHEMY_DATABASE_URI"))
        log.debug("using the connect string {0!s}".format(censor_connect_string(connect_string)))
        try:
            pool_size = self.config.get("PI_MONITORING_POOL_SIZE", 20)
            engine = create_engine(
                connect_string,
                pool_size=pool_size,
                pool_recycle=self.config.get("PI_MONITORING_POOL_RECYCLE", 600))
            log.debug("Using SQL pool size of {}".format(pool_size))
        except TypeError:
            # SQLite does not support pool_size
            engine = create_engine(connect_string)
            log.debug("Using no SQL pool_size.")
        return engine

    def add_value(self, stats_key, stats_value, timestamp, reset_values=False):
        utc_timestamp = convert_timestamp_to_utc(timestamp)
        try:
            ms = MonitoringStats(utc_timestamp, stats_key, stats_value)
            self.session.add(ms)
            self.session.commit()
            if reset_values:
                # Successfully saved the new stats entry, so remove old entries
                self.session.query(MonitoringStats).filter(and_(MonitoringStats.stats_key == stats_key,
                                                                MonitoringStats.timestamp < utc_timestamp)).delete()
                self.session.commit()
        except Exception as exx:  # pragma: no cover
            log.error("exception {0!r}".format(exx))
            log.error("DATA: {0!s} -> {1!s}".format(stats_key, stats_value))
            log.debug("{0!s}".format(traceback.format_exc()))
            self.session.rollback()

        finally:
            self.session.close()

    def delete(self, stats_key, start_timestamp, end_timestamp):
        r = None
        conditions = [MonitoringStats.stats_key == stats_key]
        if start_timestamp:
            utc_start_timestamp = convert_timestamp_to_utc(start_timestamp)
            conditions.append(MonitoringStats.timestamp >= utc_start_timestamp)
        if end_timestamp:
            utc_end_timestamp = convert_timestamp_to_utc(end_timestamp)
            conditions.append(MonitoringStats.timestamp <= utc_end_timestamp)
        try:
            r = self.session.query(MonitoringStats).filter(and_(*conditions)).delete()
            self.session.commit()
        except Exception as exx:  # pragma: no cover
            log.error("exception {0!r}".format(exx))
            log.error("could not delete statskeys {0!s}".format(stats_key))
            log.debug("{0!s}".format(traceback.format_exc()))
            self.session.rollback()

        finally:
            self.session.close()

        return r

    def get_keys(self):
        """
        Return a list of all stored keys.
        :return:
        """
        keys = []
        try:
            for monStat in self.session.query(MonitoringStats).with_entities(MonitoringStats.stats_key).distinct():
                keys.append(monStat.stats_key)
        except Exception as exx:  # pragma: no cover
            log.error("exception {0!r}".format(exx))
            log.error("could not fetch list of keys")
            log.debug("{0!s}".format(traceback.format_exc()))
            self.session.rollback()

        finally:
            self.session.close()
        return keys

    def get_values(self, stats_key, start_timestamp=None, end_timestamp=None, date_strings=False):
        values = []

        try:
            conditions = [MonitoringStats.stats_key == stats_key]
            if start_timestamp:
                utc_start_timestamp = convert_timestamp_to_utc(start_timestamp)
                conditions.append(MonitoringStats.timestamp >= utc_start_timestamp)
            if end_timestamp:
                utc_end_timestamp = convert_timestamp_to_utc(end_timestamp)
                conditions.append(MonitoringStats.timestamp <= utc_end_timestamp)
            for ms in self.session.query(MonitoringStats).filter(and_(*conditions)). \
                    order_by(MonitoringStats.timestamp.asc()):
                aware_timestamp = ms.timestamp.replace(tzinfo=tzutc())
                values.append((aware_timestamp, ms.stats_value))
        except Exception as exx:  # pragma: no cover
            log.error("exception {0!r}".format(exx))
            log.error("could not fetch list of keys")
            log.debug("{0!s}".format(traceback.format_exc()))
            self.session.rollback()

        finally:
            self.session.close()

        return values

    def get_last_value(self, stats_key):
        val = None
        try:
            s = self.session.query(MonitoringStats).filter(MonitoringStats.stats_key == stats_key). \
                order_by(MonitoringStats.timestamp.desc()).first()
            if s:
                val = s.stats_value
        except Exception as exx:  # pragma: no cover
            log.error("exception {0!r}".format(exx))
            log.error("could not fetch list of keys")
            log.debug("{0!s}".format(traceback.format_exc()))
            self.session.rollback()

        finally:
            self.session.close()

        return val



#  2023-02-22 Cornelius Kölbel <cornelius.koelbel@netknights.it>
#             initial write
#  (c) 2023 Cornelius Kölbel - cornelius.koelbel@netknights.it
#
# This code is free software; you can redistribute it and/or
# modify it under the terms of the GNU AFFERO GENERAL PUBLIC LICENSE
# License as published by the Free Software Foundation; either
# version 3 of the License, or any later version.
#
# This code is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU AFFERO GENERAL PUBLIC LICENSE for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
"""
This file contains the definition of the application specific password token class
"""

import logging
from privacyidea.lib.tokens.passwordtoken import PasswordTokenClass
from privacyidea.lib.log import log_with
from privacyidea.lib import _
from privacyidea.lib.policy import SCOPE, ACTION, GROUP
from privacyidea.api.lib.utils import getParam


TOKENINFO_KEY = "service_id"

log = logging.getLogger(__name__)


class ApplicationSpecificPasswordTokenClass(PasswordTokenClass):
    """
    This Token does use a fixed Password as the OTP value.
    In addition, the OTP PIN can be used with this token.

    This static password is tied to a certain application or service,
    making it an application specific password.
    """
    # We use an easier length of 23 for password tokens
    default_length = 23
    default_contents = 'cn'

    def __init__(self, aToken):
        PasswordTokenClass.__init__(self, aToken)
        self.set_type("applspec")

    @staticmethod
    def get_class_type():
        return "applspec"

    @staticmethod
    def get_class_prefix():
        return "ASPW"

    @staticmethod
    @log_with(log)
    def get_class_info(key=None, ret='all'):
        """
        returns a subtree of the token definition

        :param key: subsection identifier
        :type key: string
        :param ret: default return value, if nothing is found
        :type ret: user defined
        :return: subsection if key exists or user defined
        :rtype: dict or scalar
        """
        res = {'type': 'applspec',
               'title': 'Application Specific Password Token',
               'description': _('Application Specific Password: A token with a fixed password. Can be used '
                                'for certain applications or services.'),
               'init': {},
               'config': {},
               'user':  ["enroll"],
               # This tokentype is enrollable in the UI for...
               'ui_enroll': ["admin", "user"],
               'policy': {
                   SCOPE.ENROLL: {
                       ACTION.MAXTOKENUSER: {
                           'type': 'int',
                           'desc': _("The user may only have this maximum number of application specific "
                                     "password tokens assigned."),
                           'group': GROUP.TOKEN
                       },
                       ACTION.MAXACTIVETOKENUSER: {
                           'type': 'int',
                           'desc': _("The user may only have this maximum number of active application specific"
                                     " password tokens assigned."),
                           'group': GROUP.TOKEN
                       }
                   }
               },
               }
        # I don't think we need to define the lost token policies here...

        if key:
            ret = res.get(key)
        else:
            if ret == 'all':
                ret = res
        return ret

    @log_with(log, log_entry=False)
    def update(self, param):
        """
        This method is called during the initialization process.
        :param param: parameters from the token init
        :type param: dict
        :return: None
        """
        PasswordTokenClass.update(self, param)
        # In addition to the initialization from the parent class, we also need to set the service_id
        service_id = getParam(param, TOKENINFO_KEY, optional=False)
        self.add_tokeninfo(TOKENINFO_KEY, service_id)

    @property
    def service_id(self):
        r = self.get_tokeninfo()
        service_id = self.get_tokeninfo(TOKENINFO_KEY)
        return service_id

    @log_with(log)
    def use_for_authentication(self, options):
        """
        This method checks, if this token should be used for authentication.

        In this case the service_id of the token needs to match the service_id from
        the request.

        :param options: This is the option list, that basically contains the Request parameters.
        :return:
        """
        service_id = options.get(TOKENINFO_KEY)
        if not service_id:
            log.debug("The request has no {0!s}.".format(TOKENINFO_KEY))
            return False
        if not self.service_id:
            # A token could be missing the service_id
            log.debug("The token has no {0!s}.".format(TOKENINFO_KEY))
            return False
        return self.service_id.lower() == service_id.lower()



#  2018-05-07 Cornelius Kölbel <cornelius.koelbel@netknights.it>
#             Use tags in email subject.
#  2017-10-27 Cornelius Kölbel <cornelius.koelbel@netknights.it>
#             Add additional tags for notification: date, time, client_ip,
#             ua_string, ua_browser
#  2016-10-12 Cornelius Kölbel <cornelius.koelbel@netknights.it>
#             Add tokentype, tokenrealm and serial
#             Add multi and regexp
#  2016-07-18 Cornelius Kölbel <cornelius.koelbel@netknights.it>
#             Add notification conditions
#  2016-05-06 Cornelius Kölbel <cornelius.koelbel@netknights.it>
#             Initial writup
#
# License:  AGPLv3
# (c) 2016. Cornelius Kölbel
#
# This code is free software; you can redistribute it and/or
# modify it under the terms of the GNU AFFERO GENERAL PUBLIC LICENSE
# License as published by the Free Software Foundation; either
# version 3 of the License, or any later version.
#
# This code is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU AFFERO GENERAL PUBLIC LICENSE for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#
__doc__ = """This is the event handler module for user notifications.
It can be bound to each event and can perform the action:

  * sendmail: Send an email to the user/token owner
  * sendsms: We can also notify the user with an SMS.
  * savefile: Create a file which can be processed later

The module is tested in tests/test_lib_eventhandler_usernotification.py
"""

from privacyidea.lib.eventhandler.base import BaseEventHandler
from privacyidea.lib.smtpserver import send_email_identifier
from privacyidea.lib.smsprovider.SMSProvider import send_sms_identifier
from privacyidea.lib.auth import get_db_admins, get_db_admin
from privacyidea.lib.framework import get_app_config_value
from privacyidea.lib.token import get_tokens
from privacyidea.lib.smtpserver import get_smtpservers
from privacyidea.lib.smsprovider.SMSProvider import get_smsgateway
from privacyidea.lib.user import User, get_user_list, is_attribute_at_all
from privacyidea.lib.utils import create_tag_dict, to_unicode, is_true
from privacyidea.lib.crypto import get_alphanum_str
from privacyidea.lib import _
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
from email.mime.image import MIMEImage
from urllib.request import urlopen
import logging
import os
import traceback

log = logging.getLogger(__name__)

DEFAULT_BODY = """
Hello {user},

the administrator {admin}@{realm} performed the action
{action} on your token {serial}.

To check your tokens you may login to the Web UI:
{url}
"""


class NOTIFY_TYPE(object):
    """
    Allowed token owner
    """
    TOKENOWNER = "tokenowner"
    LOGGED_IN_USER = "logged_in_user"
    INTERNAL_ADMIN = "internal admin"
    ADMIN_REALM = "admin realm"
    EMAIL = "email"
    NO_REPLY_TO = ""


class UserNotificationEventHandler(BaseEventHandler):
    """
    An Eventhandler needs to return a list of actions, which it can handle.

    It also returns a list of allowed action and conditions

    It returns an identifier, which can be used in the eventhandling definitions
    """

    identifier = "UserNotification"
    description = "This eventhandler notifies the user about actions on his tokens"

    @property
    def allowed_positions(self):
        """
        This returns the allowed positions of the event handler definition.
        :return: list of allowed positions
        """
        return ["post", "pre"]

    @staticmethod
    def attach_qr(body, url_img, mimetype, serial):
        """
        Attach the QR-Code image to the email body.

        :param body: email body
        :param url_img: URL of the QR-Code image
        :param mimetype: mimetype of the email
        :param serial: serial number of the token
        :return: body with the attached QR-Code image
        """
        # get the image part of the url
        url = urlopen(url_img)  # nosec B310   # no user input
        mail_body = MIMEMultipart('related')
        mail_body.attach(MIMEText(body, mimetype))
        mail_img = MIMEImage(url.read())
        mail_img.add_header('Content-ID', '<token_image>')
        mail_img.add_header('Content-Disposition',
                            'inline; filename="{0!s}.png"'.format(serial))
        mail_body.attach(mail_img)
        body = mail_body
        return body

    @property
    def actions(self):
        """
        This method returns a dictionary of allowed actions and possible
        options in this handler module.

        :return: dict with actions
        """
        smtpserver_objs = get_smtpservers()
        smsgateway_dicts = get_smsgateway()
        smsgateways = [sms.identifier for sms in smsgateway_dicts]
        smtpservers = [s.config.identifier for s in smtpserver_objs]
        actions = {
            "sendmail": {
                "emailconfig": {
                    "type": "str",
                    "required": True,
                    "description": _("Send notification email via this email server."),
                    "value": smtpservers},
                "mimetype": {
                    "type": "str",
                    "description": _("Either send email as plain text or HTML."),
                    "value": ["plain", "html"]},
                "attach_qrcode": {
                    "type": "bool",
                    "description": _("Send QR-Code image as an attachment "
                                     "(cid URL: token_image)")},
                "subject": {
                    "type": "str",
                    "required": False,
                    "description": _("The subject of the mail that is sent.")},
                "reply_to": {
                    "type": "str",
                    "required": False,
                    "description": _("The Reply-To header in the sent email."),
                    "value": [
                        NOTIFY_TYPE.NO_REPLY_TO,
                        NOTIFY_TYPE.TOKENOWNER,
                        NOTIFY_TYPE.LOGGED_IN_USER,
                        NOTIFY_TYPE.INTERNAL_ADMIN,
                        NOTIFY_TYPE.ADMIN_REALM,
                        NOTIFY_TYPE.EMAIL]},
                "reply_to " + NOTIFY_TYPE.ADMIN_REALM: {
                    "type": "str",
                    "value": get_app_config_value("SUPERUSER_REALM", []),
                    "visibleIf": "reply_to",
                    "visibleValue": NOTIFY_TYPE.ADMIN_REALM},
                "reply_to " + NOTIFY_TYPE.INTERNAL_ADMIN: {
                    "type": "str",
                    "value": [a.username for a in
                              get_db_admins()],
                    "visibleIf": "reply_to",
                    "visibleValue":
                        NOTIFY_TYPE.INTERNAL_ADMIN},
                "reply_to " + NOTIFY_TYPE.EMAIL: {
                    "type": "str",
                    "description": _("Any email address, to which the notification "
                                     "should be sent."),
                    "visibleIf": "reply_to",
                    "visibleValue": NOTIFY_TYPE.EMAIL},
                "body": {
                    "type": "text",
                    "required": False,
                    "description": _("The template of the mail body that will be sent. It may contain the following "
                                     "tags as specified in the documentation: <code>{admin}, {realm}, {action}, "
                                     "{serial}, {url}, {user}, {givenname}, {surname}, {username}, {userrealm}, "
                                     "{tokentype}, {tokendescription}, {registrationcode}, {recipient_givenname}, "
                                     "{recipient_surname}, {googleurl_value}, {googleurl_img}, {pushurl_value}, "
                                     "{pushurl_img}, {container_url_value}, {container_url_img}, {time}, {date}, "
                                     "{client_ip}, {ua_browser}, {ua_string}, {pin}."
                                     "</code>")},
                "To": {
                    "type": "str",
                    "required": True,
                    "description": _("Send notification to this user."),
                    "value": [
                        NOTIFY_TYPE.TOKENOWNER,
                        NOTIFY_TYPE.LOGGED_IN_USER,
                        NOTIFY_TYPE.INTERNAL_ADMIN,
                        NOTIFY_TYPE.ADMIN_REALM,
                        NOTIFY_TYPE.EMAIL]},
                "To " + NOTIFY_TYPE.ADMIN_REALM: {
                    "type": "str",
                    "value": get_app_config_value("SUPERUSER_REALM", []),
                    "visibleIf": "To",
                    "visibleValue": NOTIFY_TYPE.ADMIN_REALM},
                "To " + NOTIFY_TYPE.INTERNAL_ADMIN: {
                    "type": "str",
                    "value": [a.username for a in
                              get_db_admins()],
                    "visibleIf": "To",
                    "visibleValue":
                        NOTIFY_TYPE.INTERNAL_ADMIN},
                "To " + NOTIFY_TYPE.EMAIL: {
                    "type": "str",
                    "description": _("Any email address, to which the notification "
                                     "should be sent."),
                    "visibleIf": "To",
                    "visibleValue": NOTIFY_TYPE.EMAIL}
            },
            "sendsms": {
                "smsconfig": {
                    "type": "str",
                    "required": True,
                    "description": _("Send the user notification via a "
                                     "predefined SMS gateway."),
                    "value": smsgateways},
                "body": {"type": "text",
                         "required": False,
                         "description": _("The text template of the SMS. It may contain the following tags "
                                          "as specified in the documentation: <code>{admin}, {realm}, {action}, "
                                          "{serial}, {url}, {user}, {givenname}, {surname}, {username}, {userrealm}, "
                                          "{tokentype}, {tokendescription}, {registrationcode}, {recipient_givenname}, "
                                          "{recipient_surname}, {googleurl_value}, {googleurl_img}, {pushurl_value}, "
                                          "{pushurl_img}, {container_url_value}, {container_url_img}, {time}, {date}, "
                                          "{client_ip}, {ua_browser}, {ua_string}, "
                                          "{pin}.</code>")},
                "To": {"type": "str",
                       "required": True,
                       "description": _("Send notification to this user."),
                       "value": [NOTIFY_TYPE.TOKENOWNER]}
            },
            "savefile": {
                "body": {
                    "type": "text",
                    "required": True,
                    "description": _("This is the template content of the new file. It may contain the following tags "
                                     "as specified in the documentation: <code>{admin}, {realm}, {action}, "
                                     "{serial}, {url}, {user}, {givenname}, {surname}, {username}, {userrealm}, "
                                     "{tokentype}, {tokendescription}, {registrationcode}, {recipient_givenname}, "
                                     "{recipient_surname}, {googleurl_value}, {googleurl_img}, {pushurl_value}, "
                                     "{pushurl_img}, {container_url_value}, {container_url_img}, {time}, {date}, "
                                     "{client_ip}, {ua_browser}, {ua_string}, {pin}."
                                     "</code>")},
                "filename": {
                    "type": "str",
                    "required": True,
                    "description": _("The filename of the notification. Existing files "
                                     "are overwritten. The name can contain tags as specified "
                                     "in the documentation and can also contain the tag {random}.")}
            }
        }
        return actions

    def do(self, action, options=None):
        """
        This method executes the defined action in the given event.

        :param action:
        :param options: Contains the flask parameters g, request, response
            and the handler_def configuration
        :type options: dict
        :return:
        """
        ret = True
        g = options.get("g")
        request = options.get("request")
        response = options.get("response")
        content = self._get_response_content(response)
        handler_def = options.get("handler_def")
        handler_options = handler_def.get("options", {})
        notify_type = handler_options.get("To", NOTIFY_TYPE.TOKENOWNER)
        reply_to_type = handler_options.get("reply_to")
        try:
            logged_in_user = g.logged_in_user
        except Exception:
            logged_in_user = {}

        tokenowner = self._get_tokenowner(request)
        log.debug("Executing event for action {0!r}, user {1!r}, "
                  "logged_in_user {2!r}".format(action, tokenowner,
                                                logged_in_user))

        # Determine recipient
        recipient = None
        reply_to = None

        if reply_to_type == NOTIFY_TYPE.NO_REPLY_TO:
            reply_to = ""

        elif reply_to_type == NOTIFY_TYPE.TOKENOWNER and not tokenowner.is_empty():
            reply_to = tokenowner.info.get("email")

        elif reply_to_type == NOTIFY_TYPE.INTERNAL_ADMIN:
            username = handler_options.get("reply_to " + NOTIFY_TYPE.INTERNAL_ADMIN)
            internal_admin = get_db_admin(username)
            reply_to = internal_admin.email if internal_admin else ""

        elif reply_to_type == NOTIFY_TYPE.ADMIN_REALM:
            # Adds all email addresses from a specific admin realm to the reply-to-header
            admin_realm = handler_options.get("reply_to " + NOTIFY_TYPE.ADMIN_REALM)
            attr = is_attribute_at_all()
            ulist = get_user_list({"realm": admin_realm}, custom_attributes=attr)
            # create a list of all user-emails, if the user has an email
            emails = [u.get("email") for u in ulist if u.get("email")]
            reply_to = ",".join(emails)

        elif reply_to_type == NOTIFY_TYPE.LOGGED_IN_USER:
            # Add email address from the logged in user into the reply-to header
            if logged_in_user.get("username") and not logged_in_user.get(
                    "realm"):
                # internal admins have no realm
                internal_admin = get_db_admin(logged_in_user.get("username"))
                if internal_admin:
                    reply_to = internal_admin.email if internal_admin else ""

            else:
                # Try to find the user in the specified realm
                user_obj = User(logged_in_user.get("username"),
                                logged_in_user.get("realm"))
                if user_obj:
                    reply_to = user_obj.info.get("email") if user_obj else ""

        elif reply_to_type == NOTIFY_TYPE.EMAIL:
            email = handler_options.get("reply_to " + NOTIFY_TYPE.EMAIL, "").split(",")
            reply_to = email[0]

        else:
            log.warning("Was not able to determine the email for the reply-to "
                        "header: {0!s}".format(handler_def))

        if notify_type == NOTIFY_TYPE.TOKENOWNER and not tokenowner.is_empty():
            recipient = {
                "givenname": tokenowner.info.get("givenname"),
                "surname": tokenowner.info.get("surname"),
                "username": tokenowner.login,
                "userrealm": tokenowner.realm,
                "email": tokenowner.info.get("email"),
                "mobile": tokenowner.info.get("mobile")
            }
        elif notify_type == NOTIFY_TYPE.INTERNAL_ADMIN:
            username = handler_options.get("To " + NOTIFY_TYPE.INTERNAL_ADMIN)
            internal_admin = get_db_admin(username)
            recipient = {
                "givenname": username,
                "email": internal_admin.email if internal_admin else ""
            }
        elif notify_type == NOTIFY_TYPE.ADMIN_REALM:
            # Send emails to all the users in the specified admin realm
            admin_realm = handler_options.get("To " + NOTIFY_TYPE.ADMIN_REALM)
            attr = is_attribute_at_all()
            ulist = get_user_list({"realm": admin_realm}, custom_attributes=attr)
            # create a list of all user-emails, if the user has an email
            emails = [u.get("email") for u in ulist if u.get("email")]
            recipient = {
                "givenname": "admin of realm {0!s}".format(admin_realm),
                "email": emails
            }
        elif notify_type == NOTIFY_TYPE.LOGGED_IN_USER:
            # Send notification to the logged in user
            if logged_in_user.get("username") and not logged_in_user.get(
                    "realm"):
                # internal admins have no realm
                internal_admin = get_db_admin(logged_in_user.get("username"))
                if internal_admin:
                    recipient = {
                        "givenname": logged_in_user.get("username"),
                        "email": internal_admin.email if internal_admin else ""
                    }
            else:
                # Try to find the user in the specified realm
                user_obj = User(logged_in_user.get("username"),
                                logged_in_user.get("realm"))
                if user_obj:
                    recipient = {
                        "givenname": user_obj.info.get("givenname"),
                        "surname": user_obj.info.get("surname"),
                        "email": user_obj.info.get("email"),
                        "mobile": user_obj.info.get("mobile")
                    }

        elif notify_type == NOTIFY_TYPE.EMAIL:
            email = handler_options.get("To " + NOTIFY_TYPE.EMAIL, "").split(",")
            recipient = {
                "email": email
            }
        else:
            log.warning("Was not able to determine the recipient for the user "
                        "notification: {0!s}".format(handler_def))

        if recipient or action.lower() == "savefile":
            # In case of "savefile" we do not need a recipient
            # Collect all data
            body = handler_options.get("body") or DEFAULT_BODY
            if body.startswith("file:"):  # pragma no cover
                # We read the template from the file.
                filename = body[5:]
                try:
                    with open(filename, "r", encoding="utf-8") as f:
                        body = f.read()
                except Exception as e:
                    log.warning("Failed to read email template from file {0!r}: {1!r}".format(filename, e))
                    log.debug("{0!s}".format(traceback.format_exc()))

            subject = handler_options.get("subject") or "An action was performed on your token."
            serial = (request.all_data.get("serial")
                      or content.get("detail", {}).get("serial")
                      or g.audit_object.audit_data.get("serial"))
            container_serial = request.all_data.get("container_serial")
            registrationcode = content.get("detail", {}).get("registrationcode")
            pin = content.get("detail", {}).get("pin")
            googleurl_value = content.get("detail", {}).get("googleurl",
                                                            {}).get("value")
            googleurl_img = content.get("detail", {}).get("googleurl",
                                                          {}).get("img")
            pushurl_value = content.get("detail", {}).get("pushurl",
                                                          {}).get("value")
            pushurl_img = content.get("detail", {}).get("pushurl",
                                                        {}).get("img")
            container_content = content.get("result", {}).get("value", {})
            container_url_value = container_url_img = None
            if isinstance(container_content, dict):
                container_url_value = container_content.get("container_url", {}).get("value")
                container_url_img = container_content.get("container_url", {}).get("img")
            tokentype = None
            tokendescription = None
            if serial:
                tokens = get_tokens(serial=serial)
                if tokens:
                    tokentype = tokens[0].get_tokentype()
                    tokendescription = tokens[0].token.description
            else:
                token_objects = get_tokens(user=tokenowner)
                serial = ','.join([tok.get_serial() for tok in token_objects])

            tags = create_tag_dict(logged_in_user=logged_in_user,
                                   request=request,
                                   client_ip=g.client_ip,
                                   pin=pin,
                                   googleurl_value=googleurl_value,
                                   googleurl_img=googleurl_img,
                                   pushurl_value=pushurl_value,
                                   pushurl_img=pushurl_img,
                                   recipient=recipient,
                                   tokenowner=tokenowner,
                                   serial=serial,
                                   container_serial=container_serial,
                                   tokentype=tokentype,
                                   tokendescription=tokendescription,
                                   registrationcode=registrationcode,
                                   escape_html=(action.lower() == "sendmail"
                                                and handler_options.get("mimetype", "").lower() == "html"),
                                   container_url_value=container_url_value,
                                   container_url_img=container_url_img)

            body = to_unicode(body).format(**tags)
            subject = subject.format(**tags)
            # Send notification
            if action.lower() == "sendmail":
                emailconfig = handler_options.get("emailconfig")
                mimetype = handler_options.get("mimetype", "plain")
                useremail = recipient.get("email")
                attach_qrcode = is_true(handler_options.get("attach_qrcode"))

                if attach_qrcode:
                    if googleurl_img:
                        body = self.attach_qr(body, googleurl_img, mimetype, serial)
                    elif pushurl_img:
                        body = self.attach_qr(body, pushurl_img, mimetype, serial)
                    elif container_url_img:
                        body = self.attach_qr(body, container_url_img, mimetype, serial)
                try:
                    ret = send_email_identifier(emailconfig,
                                                recipient=useremail,
                                                subject=subject, body=body,
                                                reply_to=reply_to,
                                                mimetype=mimetype)
                except Exception as exx:  # pragma: no cover
                    log.error("Failed to send email: {0!s}".format(exx))
                    self.run_details = "{0!s}".format(exx)
                    ret = False
                if ret:
                    log.info("Sent a notification email to user {0}".format(
                        recipient))
                else:
                    log.warning("Failed to send a notification email to user {0}".format(recipient))
                    self.run_details = "Failed: {0!s}.".format(useremail)

            elif action.lower() == "savefile":
                spooldir = get_app_config_value("PI_NOTIFICATION_HANDLER_SPOOLDIRECTORY",
                                                "/var/lib/privacyidea/notifications/")
                filename = handler_options.get("filename")
                random = get_alphanum_str(16)
                filename = filename.format(random=random, **tags).lstrip(os.path.sep)
                outfile = os.path.normpath(os.path.join(spooldir, filename))
                if not outfile.startswith(spooldir):
                    log.error('Cannot write outside of spooldir {0!s}!'.format(spooldir))
                else:
                    try:
                        with open(outfile, "w") as f:
                            f.write(body)
                    except Exception as err:
                        log.error("Failed to write notification file: {0!s}".format(err))

            elif action.lower() == "sendsms":
                smsconfig = handler_options.get("smsconfig")
                userphone = recipient.get("mobile")
                try:
                    ret = send_sms_identifier(smsconfig, userphone, body)
                except Exception as exx:
                    log.error("Failed to send sms: {0!s}".format(exx))
                    ret = False
                if ret:
                    log.info("Sent a notification sms to user {0}".format(
                        recipient))
                else:
                    log.warning("Failed to send a notification email to user "
                                "{0}".format(recipient))

        return ret



""" Test for the '/auth' API-endpoint """
import json
import logging

from testfixtures import log_capture
from .base import MyApiTestCase, OverrideConfigTestCase
import mock
from privacyidea.lib.config import set_privacyidea_config, SYSCONF
from privacyidea.lib.policy import (set_policy, SCOPE, ACTION, REMOTE_USER,
                                    delete_policy)
from privacyidea.lib.auth import create_db_admin
from privacyidea.lib.resolver import save_resolver, delete_resolver
from privacyidea.lib.realm import (set_realm, set_default_realm, delete_realm,
                                   get_default_realm)
from privacyidea.lib.event import set_event, delete_event
from privacyidea.lib.eventhandler.base import CONDITION
from privacyidea.lib.token import get_tokens, remove_token
from privacyidea.lib.user import User
from privacyidea.lib.utils import to_unicode, AUTH_RESPONSE
from privacyidea.config import TestingConfig
from . import ldap3mock


PWFILE = "tests/testdata/passwd-duplicate-name"


class AuthApiTestCase(MyApiTestCase):
    def test_01_auth_with_split(self):
        # By default, splitAtSign should be true
        set_privacyidea_config(SYSCONF.SPLITATSIGN, True)
        self.setUp_user_realms()
        with self.app.test_request_context('/auth',
                                           method='POST',
                                           data={"username": "cornelius",
                                                 "password": "test"}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            result = res.json.get("result")
            self.assertTrue(result.get("status"), result)
            self.assertIn('token', result.get("value"), result)
            self.assertEqual('realm1', result['value']['realm'], result)
        aentry = self.find_most_recent_audit_entry(action='POST /auth')
        self.assertEqual(aentry['action'], 'POST /auth', aentry)
        self.assertEqual(aentry['success'], 1, aentry)

        # test failed auth
        with self.app.test_request_context('/auth',
                                           method='POST',
                                           data={"username": "cornelius",
                                                 "password": "false"}):
            res = self.app.full_dispatch_request()
            self.assertEqual(401, res.status_code, res)
            result = res.json.get("result")
            self.assertFalse(result.get("status"), result)
            self.assertEqual(4031, result['error']['code'], result)
            self.assertEqual('Authentication failure. Wrong credentials',
                             result['error']['message'], result)
        aentry = self.find_most_recent_audit_entry(action='POST /auth')
        self.assertEqual(aentry['action'], 'POST /auth', aentry)
        self.assertEqual(aentry['success'], 0, aentry)

        # test with realm added to user
        with self.app.test_request_context('/auth',
                                           method='POST',
                                           data={"username": "cornelius@realm1",
                                                 "password": "test"}):
            res = self.app.full_dispatch_request()
            self.assertEqual(200, res.status_code, res)
            result = res.json.get("result")
            self.assertTrue(result.get("status"), result)
            self.assertIn('token', result.get("value"), result)
            self.assertEqual('realm1', result['value']['realm'], result)

        # test with realm added to user and unknown realm param
        with self.app.test_request_context('/auth',
                                           method='POST',
                                           data={"username": "cornelius@realm1",
                                                 "realm": "unknown",
                                                 "password": "test"}):
            res = self.app.full_dispatch_request()
            self.assertEqual(401, res.status_code, res)
            result = res.json.get("result")
            self.assertFalse(result.get("status"), result)
            self.assertEqual(4031, result['error']['code'], result)
            self.assertEqual('Authentication failure. Unknown realm: unknown.',
                             result['error']['message'], result)

        # test with realm parameter
        with self.app.test_request_context('/auth',
                                           method='POST',
                                           data={"username": "cornelius",
                                                 "realm": "realm1",
                                                 "password": "test"}):
            res = self.app.full_dispatch_request()
            self.assertEqual(200, res.status_code, res)
            result = res.json.get("result")
            self.assertTrue(result.get("status"), result)
            self.assertIn('token', result.get("value"), result)
            self.assertEqual('realm1', result['value']['realm'], result)

        # test with broken realm parameter
        with self.app.test_request_context('/auth',
                                           method='POST',
                                           data={"username": "cornelius",
                                                 "realm": "unknown",
                                                 "password": "test"}):
            res = self.app.full_dispatch_request()
            self.assertEqual(401, res.status_code, res)
            result = res.json.get("result")
            self.assertFalse(result.get("status"), result)
            self.assertEqual(4031, result['error']['code'], result)
            self.assertEqual('Authentication failure. Unknown realm: unknown.',
                             result['error']['message'], result)

        # test with empty realm parameter
        with self.app.test_request_context('/auth',
                                           method='POST',
                                           data={"username": "cornelius",
                                                 "realm": "",
                                                 "password": "test"}):
            res = self.app.full_dispatch_request()
            self.assertEqual(200, res.status_code, res)
            result = res.json.get("result")
            self.assertTrue(result.get("status"), result)
            self.assertIn('token', result.get("value"), result)
            # realm1 should be the default realm
            self.assertEqual('realm1', result['value']['realm'], result)

        # TODO: Add test with empty realm parameter and user@realm not in default realm

        # test with realm parameter and wrong realm added to user
        with mock.patch("logging.Logger.error") as mock_log:
            with self.app.test_request_context('/auth',
                                               method='POST',
                                               data={"username": "cornelius@unknown",
                                                     "realm": "realm1",
                                                     "password": "test"}):
                res = self.app.full_dispatch_request()
                self.assertEqual(401, res.status_code, res)
                result = res.json.get("result")
                self.assertFalse(result.get("status"), result)
            expected = "The user User(login='cornelius@unknown', " \
                       "realm='realm1', resolver='') exists in NO resolver."
            mock_log.assert_called_once_with(expected)

        # test with wrong realm parameter and wrong realm added to user
        with self.app.test_request_context('/auth',
                                           method='POST',
                                           data={"username": "cornelius@unknown",
                                                 "realm": "anotherunknown",
                                                 "password": "test"}):
            res = self.app.full_dispatch_request()
            self.assertEqual(401, res.status_code, res)
            result = res.json.get("result")
            self.assertFalse(result.get("status"), result)
            self.assertEqual(4031, result['error']['code'], result)
            self.assertEqual('Authentication failure. Unknown realm: anotherunknown.',
                             result['error']['message'], result)

        # Now we take it up one notch and add another resolver
        self.setUp_user_realm3()
        # The selfservice user does not exist in realm3
        with self.app.test_request_context('/auth',
                                           method='POST',
                                           data={"username": "selfservice@realm3",
                                                 "realm": "realm1",
                                                 "password": "test"}):
            res = self.app.full_dispatch_request()
            self.assertEqual(200, res.status_code, res)
            result = res.json.get("result")
            self.assertTrue(result.get("status"), result)
            self.assertIn('token', result.get("value"), result)
            self.assertEqual('realm1', result['value']['realm'], result)

        # and the other way round
        with mock.patch("logging.Logger.error") as mock_log:
            with self.app.test_request_context('/auth',
                                               method='POST',
                                               data={"username": "selfservice@realm1",
                                                     "realm": "realm3",
                                                     "password": "test"}):
                res = self.app.full_dispatch_request()
                self.assertEqual(401, res.status_code, res)
                result = res.json.get("result")
                self.assertFalse(result.get("status"), result)
                self.assertEqual(4031, result['error']['code'], result)
                self.assertEqual('Authentication failure. Wrong credentials',
                                 result['error']['message'], result)
            # the realm will be split from the login name
            expected = "The user User(login='selfservice', " \
                       "realm='realm3', resolver='') exists in NO resolver."
            mock_log.assert_called_once_with(expected)

    # And now we do all of the above without the splitAtSign setting
    def test_02_auth_without_split(self):
        set_privacyidea_config(SYSCONF.SPLITATSIGN, False)
        self.setUp_user_realms()
        with self.app.test_request_context('/auth',
                                           method='POST',
                                           data={"username": "cornelius",
                                                 "password": "test"}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            result = res.json.get("result")
            self.assertTrue(result.get("status"), result)
            self.assertIn('token', result.get("value"), result)
            self.assertEqual('realm1', result['value']['realm'], result)

        # test failed auth wrong password
        with self.app.test_request_context('/auth',
                                           method='POST',
                                           data={"username": "cornelius",
                                                 "password": "false"}):
            res = self.app.full_dispatch_request()
            self.assertEqual(401, res.status_code, res)
            result = res.json.get("result")
            self.assertFalse(result.get("status"), result)
            self.assertEqual(4031, result['error']['code'], result)
            self.assertEqual('Authentication failure. Wrong credentials',
                             result['error']['message'], result)

        # test failed auth no password
        with self.app.test_request_context('/auth',
                                           method='POST',
                                           data={"username": "cornelius"}):
            res = self.app.full_dispatch_request()
            self.assertEqual(401, res.status_code, res)
            result = res.json.get("result")
            self.assertFalse(result.get("status"), result)
            self.assertEqual(4031, result['error']['code'], result)
            self.assertEqual('Authentication failure. Wrong credentials',
                             result['error']['message'], result)

        # test with realm added to user. This fails since we do not split
        with mock.patch("logging.Logger.error") as mock_log:
            with self.app.test_request_context('/auth',
                                               method='POST',
                                               data={"username": "cornelius@realm1",
                                                     "password": "test"}):
                res = self.app.full_dispatch_request()
                self.assertEqual(401, res.status_code, res)
                result = res.json.get("result")
                self.assertFalse(result.get("status"), result)
                self.assertEqual(4031, result['error']['code'], result)
                self.assertEqual('Authentication failure. Wrong credentials',
                                 result['error']['message'], result)
            expected = "The user User(login='cornelius@realm1', " \
                       "realm='realm1', resolver='') exists in NO resolver."
            mock_log.assert_called_once_with(expected)

        # test with realm added to user and unknown realm param
        with self.app.test_request_context('/auth',
                                           method='POST',
                                           data={"username": "cornelius@realm1",
                                                 "realm": "unknown",
                                                 "password": "test"}):
            res = self.app.full_dispatch_request()
            self.assertEqual(401, res.status_code, res)
            result = res.json.get("result")
            self.assertFalse(result.get("status"), result)
            self.assertEqual(4031, result['error']['code'], result)
            self.assertEqual('Authentication failure. Unknown realm: unknown.',
                             result['error']['message'], result)

        # test with realm parameter
        with self.app.test_request_context('/auth',
                                           method='POST',
                                           data={"username": "cornelius",
                                                 "realm": "realm1",
                                                 "password": "test"}):
            res = self.app.full_dispatch_request()
            self.assertEqual(200, res.status_code, res)
            result = res.json.get("result")
            self.assertTrue(result.get("status"), result)
            self.assertIn('token', result.get("value"), result)
            self.assertEqual('realm1', result['value']['realm'], result)

        # test with broken realm parameter
        with self.app.test_request_context('/auth',
                                           method='POST',
                                           data={"username": "cornelius",
                                                 "realm": "unknown",
                                                 "password": "test"}):
            res = self.app.full_dispatch_request()
            self.assertEqual(401, res.status_code, res)
            result = res.json.get("result")
            self.assertFalse(result.get("status"), result)
            self.assertEqual(4031, result['error']['code'], result)
            self.assertEqual('Authentication failure. Unknown realm: unknown.',
                             result['error']['message'], result)

        # test with empty realm parameter
        with self.app.test_request_context('/auth',
                                           method='POST',
                                           data={"username": "cornelius",
                                                 "realm": "",
                                                 "password": "test"}):
            res = self.app.full_dispatch_request()
            self.assertEqual(200, res.status_code, res)
            result = res.json.get("result")
            self.assertTrue(result.get("status"), result)
            self.assertIn('token', result.get("value"), result)
            # realm1 should be the default realm
            self.assertEqual('realm1', result['value']['realm'], result)

        # test with realm parameter and wrong realm added to user
        with mock.patch("logging.Logger.error") as mock_log:
            with self.app.test_request_context('/auth',
                                               method='POST',
                                               data={"username": "cornelius@unknown",
                                                     "realm": "realm1",
                                                     "password": "test"}):
                res = self.app.full_dispatch_request()
                self.assertEqual(401, res.status_code, res)
                result = res.json.get("result")
                self.assertFalse(result.get("status"), result)
            expected = "The user User(login='cornelius@unknown', " \
                       "realm='realm1', resolver='') exists in NO resolver."
            mock_log.assert_called_once_with(expected)

        # test with wrong realm parameter and wrong realm added to user
        with self.app.test_request_context('/auth',
                                           method='POST',
                                           data={"username": "cornelius@unknown",
                                                 "realm": "anotherunknown",
                                                 "password": "test"}):
            res = self.app.full_dispatch_request()
            self.assertEqual(401, res.status_code, res)
            result = res.json.get("result")
            self.assertFalse(result.get("status"), result)
            self.assertEqual(4031, result['error']['code'], result)
            self.assertEqual('Authentication failure. Unknown realm: anotherunknown.',
                             result['error']['message'], result)

        # Now we take it up one notch and add another resolver
        self.setUp_user_realm3()
        # The selfservice user does not exist in realm3
        with mock.patch("logging.Logger.error") as mock_log:
            with self.app.test_request_context('/auth',
                                               method='POST',
                                               data={"username": "selfservice@realm3",
                                                     "realm": "realm1",
                                                     "password": "test"}):
                res = self.app.full_dispatch_request()
                self.assertEqual(401, res.status_code, res)
                result = res.json.get("result")
                self.assertFalse(result.get("status"), result)
                self.assertEqual(4031, result['error']['code'], result)
            expected = "The user User(login='selfservice@realm3', " \
                       "realm='realm1', resolver='') exists in NO resolver."
            mock_log.assert_called_once_with(expected)

        # and the other way round
        with mock.patch("logging.Logger.error") as mock_log:
            with self.app.test_request_context('/auth',
                                               method='POST',
                                               data={"username": "selfservice@realm1",
                                                     "realm": "realm3",
                                                     "password": "test"}):
                res = self.app.full_dispatch_request()
                self.assertEqual(401, res.status_code, res)
                result = res.json.get("result")
                self.assertFalse(result.get("status"), result)
                self.assertEqual(4031, result['error']['code'], result)
                self.assertEqual('Authentication failure. Wrong credentials',
                                 result['error']['message'], result)
            # the realm will be split from the login name
            expected = "The user User(login='selfservice@realm1', " \
                       "realm='realm3', resolver='') exists in NO resolver."
            mock_log.assert_called_once_with(expected)

        set_privacyidea_config(SYSCONF.SPLITATSIGN, True)

    def test_03_admin_auth(self):
        set_privacyidea_config(SYSCONF.SPLITATSIGN, True)
        # check that the default admin works
        with self.app.test_request_context('/auth',
                                           method='POST',
                                           data={"username": "testadmin",
                                                 "password": "testpw"}):
            res = self.app.full_dispatch_request()
            self.assertEqual(200, res.status_code, res)
            result = res.json.get("result")
            self.assertTrue(result.get("status"), result)
            self.assertIn('token', result.get("value"), result)
            # role should be 'admin'
            self.assertEqual('admin', result['value']['role'], result)

        # add an admin with an '@' in the login name
        create_db_admin('super@intern', password='testing')
        # as long as the part after the '@' does not resemble an existing realm,
        # this should work with 'spltAtSign' set to True
        with self.app.test_request_context('/auth',
                                           method='POST',
                                           data={"username": "super@intern",
                                                 "password": "testing"}):
            res = self.app.full_dispatch_request()
            self.assertEqual(200, res.status_code, res)
            result = res.json.get("result")
            self.assertTrue(result.get("status"), result)
            self.assertIn('token', result.get("value"), result)
            # role should be 'admin'
            self.assertEqual('admin', result['value']['role'], result)

        # both admin logins should also work with 'splitAtSign' set to False
        set_privacyidea_config(SYSCONF.SPLITATSIGN, False)
        with self.app.test_request_context('/auth',
                                           method='POST',
                                           data={"username": "testadmin",
                                                 "password": "testpw"}):
            res = self.app.full_dispatch_request()
            self.assertEqual(200, res.status_code, res)
            result = res.json.get("result")
            self.assertTrue(result.get("status"), result)
            self.assertIn('token', result.get("value"), result)
            # role should be 'admin'
            self.assertEqual('admin', result['value']['role'], result)

        with self.app.test_request_context('/auth',
                                           method='POST',
                                           data={"username": "super@intern",
                                                 "password": "testing"}):
            res = self.app.full_dispatch_request()
            self.assertEqual(200, res.status_code, res)
            result = res.json.get("result")
            self.assertTrue(result.get("status"), result)
            self.assertIn('token', result.get("value"), result)
            # role should be 'admin'
            self.assertEqual('admin', result['value']['role'], result)

        # reset 'splitAtSign' to default value
        set_privacyidea_config(SYSCONF.SPLITATSIGN, True)

    def test_04_remote_user_auth(self):
        self.setUp_user_realms()
        # first check that without a remote_user policy the login fails
        with self.app.test_request_context('/auth',
                                           method='POST',
                                           data={"username": "cornelius"},
                                           environ_base={"REMOTE_USER": "cornelius"}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 401, res)
            result = res.json.get("result")
            self.assertFalse(result.get("status"), result)
            self.assertEqual(result.get("error").get('code'), 4031, result)
            self.assertEqual(result.get("error").get('message'),
                             'Authentication failure. Wrong credentials', result)
        aentry = self.find_most_recent_audit_entry(action='POST /auth')
        self.assertEqual(aentry['action'], 'POST /auth', aentry)
        self.assertEqual(aentry['success'], 0, aentry)

        # now check that with a disabled remote_user policy the login fails
        set_policy(name="remote", scope=SCOPE.WEBUI,
                   action="{0!s}={1!s}".format(ACTION.REMOTE_USER, REMOTE_USER.DISABLE))
        with self.app.test_request_context('/auth',
                                           method='POST',
                                           data={"username": "cornelius"},
                                           environ_base={"REMOTE_USER": "cornelius"}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 401, res)
            result = res.json.get("result")
            self.assertFalse(result.get("status"), result)
            self.assertEqual(result.get("error").get('code'), 4031, result)
            self.assertEqual(result.get("error").get('message'),
                             'Authentication failure. Wrong credentials', result)
        aentry = self.find_most_recent_audit_entry(action='POST /auth')
        self.assertEqual(aentry['action'], 'POST /auth', aentry)
        self.assertEqual(aentry['success'], 0, aentry)
        self.assertEqual(aentry['policies'], 'remote', aentry)

        # And now check that with an enabled remote_user policy the login succeeds
        set_policy(name="remote", scope=SCOPE.WEBUI,
                   action="{0!s}={1!s}".format(ACTION.REMOTE_USER, REMOTE_USER.ACTIVE))
        with self.app.test_request_context('/auth',
                                           method='POST',
                                           data={"username": "cornelius"},
                                           environ_base={"REMOTE_USER": "cornelius"}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 200, res)
            result = res.json.get("result")
            self.assertTrue(result.get("status"), result)
            self.assertEqual(result.get("value").get('role'), 'user', result)
        aentry = self.find_most_recent_audit_entry(action='POST /auth')
        self.assertEqual(aentry['action'], 'POST /auth', aentry)
        self.assertEqual(aentry['success'], 1, aentry)
        self.assertEqual(aentry['policies'], 'remote', aentry)

        # check that a remote user with "@" works as well
        set_policy(name="remote", scope=SCOPE.WEBUI, realm=self.realm1,
                   action="{0!s}={1!s}".format(ACTION.REMOTE_USER, REMOTE_USER.ACTIVE))
        with self.app.test_request_context('/auth',
                                           method='POST',
                                           data={"username": "cornelius@realm1"},
                                           environ_base={"REMOTE_USER": "cornelius@realm1"}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 200, res)
            result = res.json.get("result")
            self.assertTrue(result.get("status"), result)
            self.assertEqual(result.get("value").get('role'), 'user', result)
        aentry = self.find_most_recent_audit_entry(action='POST /auth')
        self.assertEqual(aentry['action'], 'POST /auth', aentry)
        self.assertEqual(aentry['success'], 1, aentry)
        self.assertEqual(aentry['policies'], 'remote', aentry)

        # check that the policy remote_user=force passes the necessary hidden tag to the
        # login window
        set_policy(name="remote", scope=SCOPE.WEBUI, realm=self.realm1,
                   action="{0!s}={1!s}".format(ACTION.REMOTE_USER, REMOTE_USER.FORCE))
        with self.app.test_request_context('/',
                                           method='GET',
                                           environ_base={"REMOTE_USER": "cornelius@realm1"}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 200, res)
            # The login page contains the info about force remote_user, which will hide the
            # "login with credentials" button.
            self.assertIn('input type=hidden id=FORCE_REMOTE_USER value="True"', to_unicode(res.data))

        # bind the remote user policy to an unknown realm
        set_policy(name="remote", scope=SCOPE.WEBUI, realm='unknown',
                   action="{0!s}={1!s}".format(ACTION.REMOTE_USER, REMOTE_USER.ACTIVE))
        with self.app.test_request_context('/auth',
                                           method='POST',
                                           data={"username": "cornelius@realm1"},
                                           environ_base={"REMOTE_USER": "cornelius@realm1"}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 401, res)
            result = res.json.get("result")
            self.assertFalse(result.get("status"), result)
        aentry = self.find_most_recent_audit_entry(action='POST /auth')
        self.assertEqual(aentry['action'], 'POST /auth', aentry)
        self.assertEqual(aentry['success'], 0, aentry)
        self.assertEqual(None, aentry['policies'], aentry)

        # check split@sign is working correctly
        set_policy(name="remote", scope=SCOPE.WEBUI, realm=self.realm1,
                   action="{0!s}={1!s}".format(ACTION.REMOTE_USER, REMOTE_USER.ACTIVE))
        set_privacyidea_config(SYSCONF.SPLITATSIGN, False)
        with self.app.test_request_context('/auth',
                                           method='POST',
                                           data={"username": "user@test"},
                                           environ_base={"REMOTE_USER": "user@test"}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 200, res)
            result = res.json.get("result")
            self.assertTrue(result.get("status"), result)
            self.assertEqual(result.get("value").get('role'), 'user', result)
        aentry = self.find_most_recent_audit_entry(action='POST /auth')
        self.assertEqual(aentry['action'], 'POST /auth', aentry)
        self.assertEqual(aentry['success'], 1, aentry)
        self.assertEqual(aentry['policies'], 'remote', aentry)

        delete_policy(name='remote')
        set_privacyidea_config(SYSCONF.SPLITATSIGN, True)

    @ldap3mock.activate
    def test_05_local_admin_with_failing_resolver(self):
        ldap3mock.setLDAPDirectory([])
        # define, that we want to get an exception
        ldap3mock.set_exception()
        # Create an LDAP Realm as default realm
        params = {'LDAPURI': 'ldap://localhost',
                  'LDAPBASE': 'o=test',
                  'BINDDN': 'cn=manager,ou=example,o=test',
                  'BINDPW': 'ldaptest',
                  'LOGINNAMEATTRIBUTE': 'cn',
                  'LDAPSEARCHFILTER': '(cn=*)',
                  'USERINFO': '{ "username": "cn",'
                              '"phone" : "telephoneNumber", '
                              '"mobile" : "mobile"'
                              ', "email" : "mail", '
                              '"surname" : "sn", '
                              '"givenname" : "givenName" }',
                  'UIDTYPE': 'DN',
                  "resolver": "ldap1",
                  "type": "ldapresolver"}
        save_resolver(params)
        set_realm("ldap1", [{'name': "ldap1"}])
        set_default_realm("ldap1")

        # Try to log in as internal admin with a failing LDAP resolver
        with mock.patch("logging.Logger.warning") as mock_log:
            with self.app.test_request_context('/auth',
                                               method='POST',
                                               data={"username": "testadmin",
                                                     "password": "testpw"}):
                res = self.app.full_dispatch_request()
                self.assertEqual(200, res.status_code, res)
                result = res.json.get("result")
                self.assertTrue(result.get("status"), result)
                self.assertIn('token', result.get("value"), result)
                # role should be 'admin'
                self.assertEqual('admin', result['value']['role'], result)
                mock_log.assert_called_once_with("Problem resolving user testadmin in "
                                                 "realm ldap1: ERR907: Error performing "
                                                 "bind operation: LDAP request failed.")

        delete_realm("ldap1")
        delete_resolver("ldap1")
        ldap3mock.set_exception(False)

    def test_06_auth_user_not_in_defrealm(self):
        self.setUp_user_realms()
        self.setUp_user_realm3()
        set_default_realm(self.realm3)
        # check that realm3 is the default realm
        self.assertEqual(self.realm3, get_default_realm())
        # authentication with "@realm1" works
        with self.app.test_request_context('/auth',
                                           method='POST',
                                           data={"username": "selfservice@realm1",
                                                 "password": "test"}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 200, res)
            result = res.json.get("result")
            self.assertTrue(result.get("status"), result)
            self.assertIn('token', result.get("value"), result)
            self.assertEqual('realm1', result['value']['realm'], result)

        # while authentication without a realm fails (user "selfservice"
        # doesn't exist in default realm "realm3")
        with self.app.test_request_context('/auth',
                                           method='POST',
                                           data={"username": "selfservice",
                                                 "password": "test"}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 401, res)
            result = res.json.get("result")
            self.assertFalse(result.get("status"), result)

        # A given empty realm parameter should not change the realm of the user
        with self.app.test_request_context('/auth',
                                           method='POST',
                                           data={"username": "selfservice@realm1",
                                                 "password": "test",
                                                 "realm": ""}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 200, res)
            result = res.json.get("result")
            self.assertTrue(result.get("status"), result)
            self.assertIn('token', result.get("value"), result)
            self.assertEqual('realm1', result['value']['realm'], result)

    def test_07_user_not_in_userstore(self):
        # If a user can not be found in the userstore we always get the response
        # "Wrong Credentials"
        self.setUp_user_realms()
        set_default_realm(self.realm1)

        # user authenticates against userstore but user does not exist
        with self.app.test_request_context('/auth',
                                           method='POST',
                                           data={"username": "user-really-does-not-exist",
                                                 "password": "test"}):
            res = self.app.full_dispatch_request()
            self.assertEqual(401, res.status_code, res)
            result = res.json.get("result")
            self.assertFalse(result.get("status"), result)
            error = result.get("error")
            self.assertEqual(4031, error.get("code"))
            self.assertEqual("Authentication failure. Wrong credentials", error.get("message"))

        # set a policy to authenticate against privacyIDEA
        set_policy("piLogin", scope=SCOPE.WEBUI, action="{0!s}=privacyIDEA".format(ACTION.LOGINMODE))

        # user authenticates against privacyidea but user does not exist
        with self.app.test_request_context('/auth',
                                           method='POST',
                                           data={"username": "user-really-does-not-exist",
                                                 "password": "test"}):
            res = self.app.full_dispatch_request()
            self.assertEqual(401, res.status_code, res)
            result = res.json.get("result")
            self.assertFalse(result.get("status"), result)
            error = result.get("error")
            self.assertEqual(4031, error.get("code"))
            self.assertEqual("Authentication failure. Wrong credentials", error.get("message"))

        # cleanup
        delete_policy("piLogin")
        delete_realm(self.realm1)
        delete_resolver(self.resolvername1)


class AdminFromUserstore(OverrideConfigTestCase):
    class Config(TestingConfig):
        SUPERUSER_REALM = ["realm1"]

    def test_01_admin_from_userstore(self):
        self.setUp_user_realms()
        # login as an admin user from userstore
        with self.app.test_request_context('/auth',
                                           method='POST',
                                           data={"username": "cornelius@realm1",
                                                 "password": "test"}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 200, res)
            result = res.json.get("result")
            self.assertTrue(result.get("status"), result)
            self.assertIn('token', result.get('value'), result)
            self.assertEqual(result.get("value").get('role'), 'admin', result)
            # if no admin policy is set, the user should have all admin rights
            self.assertIn(ACTION.DELETE, result.get('value').get('rights'), result)

        # Now test with a helpdesk policy for the admin realm
        set_policy(name='helpdesk', scope=SCOPE.ADMIN, adminrealm=self.realm1,
                   action=ACTION.DISABLE)
        with self.app.test_request_context('/auth',
                                           method='POST',
                                           data={"username": "cornelius@Realm1",
                                                 "password": "test"}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 200, res)
            result = res.json.get("result")
            self.assertTrue(result.get("status"), result)
            self.assertIn('token', result.get('value'), result)
            self.assertEqual(result.get("value").get('role'), 'admin', result)
            # check that the appropriate rights are set/unset
            self.assertNotIn(ACTION.DELETE, result.get('value').get('rights'), result)
            self.assertIn(ACTION.DISABLE, result.get('value').get('rights'), result)
        delete_policy(name='helpdesk')


class DuplicateUserApiTestCase(MyApiTestCase):

    def test_01_admin_and_user_same_name(self):
        # Test the logging, if admin and user have the same name (testadmin/testpw)
        # Now create a default realm, that contains the used "testadmin"
        rid = save_resolver({"resolver": self.resolvername1,
                             "type": "passwdresolver",
                             "fileName": PWFILE})
        self.assertTrue(rid > 0, rid)

        (added, failed) = set_realm(self.realm1,
                                    [{'name': self.resolvername1}])
        self.assertTrue(len(failed) == 0)
        self.assertTrue(len(added) == 1)

        set_default_realm(self.realm1)

        # If the admin logs in, everything is fine
        with mock.patch("logging.Logger.info") as mock_log:
            with self.app.test_request_context('/auth',
                                               method='POST',
                                               data={"username": "testadmin",
                                                     "password": "testpw"}):
                res = self.app.full_dispatch_request()
                self.assertEqual(200, res.status_code, res)
                result = res.json.get("result")
                self.assertTrue(result.get("status"), result)
                self.assertIn('token', result.get("value"), result)
                # role should be 'admin'
                self.assertEqual('admin', result['value']['role'], result)
            mock_log.assert_any_call("Local admin 'testadmin' successfully logged in.")

        # If a user logs in, with the same name as the admin, this event is logged in warning
        with mock.patch("logging.Logger.warning") as mock_log:
            with self.app.test_request_context('/auth',
                                               method='POST',
                                               data={"username": "testadmin",
                                                     "password": "test"}):
                res = self.app.full_dispatch_request()
                self.assertEqual(200, res.status_code, res)
                result = res.json.get("result")
                self.assertTrue(result.get("status"), result)
                self.assertIn('token', result.get("value"), result)
                # role should be 'user'
                self.assertEqual('user', result['value']['role'], result)
            # check if we have this log entry
            mock_log.assert_called_with("A user 'testadmin' exists as local admin and as user "
                                        "in your default realm!")

        # Check that a wrong/missing password doesn't trigger the warning in the log
        with mock.patch("logging.Logger.warning") as mock_log:
            with self.app.test_request_context('/auth',
                                               method='POST',
                                               data={"username": "testadmin",
                                                     "password": "unknown"}):
                res = self.app.full_dispatch_request()
                self.assertEqual(401, res.status_code, res)
                result = res.json.get("result")
                self.assertFalse(result.get("status"), result)
                self.assertEqual(result.get("error").get('code'), 4031, result)
                self.assertEqual(result.get("error").get('message'),
                                 'Authentication failure. Wrong credentials', result)
            # check if we have the correct log entry
            mock_log.assert_called_with("user uid 1004 failed to authenticate")


class EventHandlerTest(MyApiTestCase):

    def test_01_pre_eventhandlers(self):
        # This test create an HOTP token with C/R with a pre-event handler
        # and the user uses this HOTP token to directly login to /auth

        # Setup realm
        rid = save_resolver({"resolver": self.resolvername1,
                             "type": "passwdresolver",
                             "fileName": PWFILE})
        self.assertTrue(rid > 0, rid)

        (added, failed) = set_realm(self.realm1,
                                    [{'name': self.resolvername1}])
        self.assertTrue(len(failed) == 0)
        self.assertTrue(len(added) == 1)

        set_default_realm(self.realm1)

        # set a policy to authenticate against privacyIDEA
        set_policy("piLogin", scope=SCOPE.WEBUI, action="{0!s}=privacyIDEA".format(ACTION.LOGINMODE))
        # set a policy to for otppin=userstore
        set_policy("otppin", scope=SCOPE.AUTH, action="{0!s}=userstore".format(ACTION.OTPPIN))
        # Set a policy to do C/R with HOTP tokens
        set_policy("crhotp", scope=SCOPE.AUTH, action="{0!s}=hotp".format(ACTION.CHALLENGERESPONSE))

        # Create an event handler, that creates HOTP token on /auth with default OTP key
        # TODO: this is probably a bad test case since enrolling an HOTP-Token
        #  during `/auth` does not return the generated secret/QR-code.
        #  For now this is useful to test the `/auth` endpoint with challenge-response.
        eid = set_event("createtoken", event=["auth"], handlermodule="Token",
                        action="enroll", position="pre",
                        conditions={CONDITION.USER_TOKEN_NUMBER: 0},
                        options={"tokentype": "hotp", "user": "1",
                                 "additional_params": {
                                     'otpkey': self.otpkey,
                                     # We need to set genkey=0, otherwise the Tokenhandler will
                                     # generate a random otpkey
                                     'genkey': 0}})
        # cleanup tokens
        remove_token(user=User("someuser", self.realm1))

        # user tries to log in with his userstore password and gets a transaction_id
        with self.app.test_request_context('/auth',
                                           method='POST',
                                           data={"username": "someuser",
                                                 "password": "test"}):
            res = self.app.full_dispatch_request()
            self.assertEqual(200, res.status_code, res)
            result = res.json.get("result")
            self.assertTrue(result.get("status"), result)
            detail = res.json.get("detail")
            self.assertEqual("please enter otp: ", detail.get("message"))
            transaction_id = detail.get("transaction_id")
            # In case of a challenge request, make sure no user data is spilled
            self.assertNotIn("someuser", json.dumps(res.json), res.json)
            self.assertFalse(result.get("value"), result)
            self.assertEqual(result.get("authentication"), AUTH_RESPONSE.CHALLENGE, result)

        # Check if the token was enrolled
        toks = get_tokens(user=User("someuser", self.realm1))
        self.assertEqual(len(toks), 1)
        self.assertEqual(toks[0].token.tokentype, "hotp")
        serial = toks[0].token.serial
        # Check if the correct otpkey was used
        hotptoken = toks[0]
        r = hotptoken.check_otp(self.valid_otp_values[1])
        self.assertTrue(r >= 0)

        # Now the user logs in with the second step of C/R with OTP value of new token
        with self.app.test_request_context('/auth',
                                           method='POST',
                                           data={"username": "someuser",
                                                 "transaction_id": transaction_id,
                                                 "password": self.valid_otp_values[2]}):
            res = self.app.full_dispatch_request()
            self.assertEqual(200, res.status_code, res)
            result = res.json.get("result")
            self.assertTrue(result.get("status"))
            self.assertEqual(result.get("authentication"), AUTH_RESPONSE.ACCEPT, result)
            self.assertTrue(result.get("value"))

        # Check that there is still only one token
        toks = get_tokens(user=User("someuser", self.realm1))
        self.assertEqual(len(toks), 1)
        self.assertEqual(toks[0].token.tokentype, "hotp")
        self.assertEqual(serial, toks[0].token.serial)

        # cleanup
        delete_policy("piLogin")
        delete_policy("otppin")
        delete_policy("crhotp")
        delete_event(eid)
        remove_token(hotptoken.token.serial)

    @log_capture
    def test_02_post_eventhandler(self, capture):
        self.setUp_user_realms()
        # Create an event handler, that creates HOTP token on /auth with default OTP key
        eid = set_event("post_event_log", event=["auth"], handlermodule="Logging",
                        action="logging", position="post",
                        options={"level": logging.INFO,
                                 "message": "User: {user} Event: {action}"})

        with self.app.test_request_context('/auth',
                                           method='POST',
                                           data={"username": "someuser",
                                                 "password": "test"}):
            res = self.app.full_dispatch_request()
            self.assertEqual(200, res.status_code, res)
            result = res.json.get("result")
            self.assertTrue(result.get("status"), result)

        capture.check_present(
            ('pi-eventlogger', 'INFO',
             'User: someuser Event: /auth')
        )

        delete_event(eid)
